/***************************************************************************
 *                                                                         *
 *   Project: mfx receiver running on a PIC16F1705                         *
 *   File:    main.c                                                       *
 *   (c) 2019 - 2022 Rainer Müller                                         *
 *                                                                         *
 ***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// PIC16F1705 Configuration Bit Settings

// CONFIG1
#pragma config FOSC = INTOSC    // Oscillator Selection Bits (INTOSC oscillator: I/O function on CLKIN pin)
#pragma config WDTE = OFF       // Watchdog Timer Enable (WDT disabled)
#pragma config PWRTE = OFF      // Power-up Timer Enable (PWRT disabled)
#pragma config MCLRE = ON       // MCLR Pin Function Select (MCLR/VPP pin function is MCLR)
#pragma config CP = OFF         // Flash Program Memory Code Protection (Program memory code protection is disabled)
#pragma config BOREN = ON       // Brown-out Reset Enable (Brown-out Reset enabled)
#pragma config CLKOUTEN = ON    // Clock Out Enable (CLKOUT function is enabled on the CLKOUT pin)
#pragma config IESO = OFF       // Internal/External Switchover Mode (Internal/External Switchover Mode is disabled)
#pragma config FCMEN = ON       // Fail-Safe Clock Monitor Enable (Fail-Safe Clock Monitor is enabled)

// CONFIG2
#pragma config WRT = OFF        // Flash Memory Self-Write Protection (Write protection off)
#pragma config PPS1WAY = OFF    // Peripheral Pin Select one-way control (The PPSLOCK bit can be set and cleared repeatedly by software)
#pragma config ZCDDIS = ON      // Zero-cross detect disable (Zero-cross detect circuit is disabled at POR)
#pragma config PLLEN = OFF      // Phase Lock Loop enable (4x PLL is enabled when software sets the SPLLEN bit)
#pragma config STVREN = ON      // Stack Overflow/Underflow Reset Enable (Stack Overflow or Underflow will cause a Reset)
#pragma config BORV = LO        // Brown-out Reset Voltage Selection (Brown-out Reset Voltage (Vbor), low trip point selected.)
#pragma config LPBOR = OFF      // Low-Power Brown Out Reset (Low-Power BOR is disabled)
#pragma config LVP = ON         // Low-Voltage Programming Enable (Low-voltage programming enabled)

#include <xc.h>
#include <stdint.h>


#define CARR_STAB	60
#define CARR_LOST	20

enum { IDLE, STAB };		// detection phases

uint8_t  askval, carrier, detectstep, spirx, spitx, loops;
uint16_t level = 0;
//uint8_t  testbuf[256];

const uint8_t phasetab[] = {
	0,	// 0000
	0,	// 0001
	0,	// 0010
	1,	// 0011 -> 225°
	0,	// 0100
	0,	// 0101
	1,	// 0110 -> 315°
	0,	// 0111
	0,	// 1000
	1,	// 1001 -> 135°
	0,  // 1010
	0,	// 1011
	1,	// 1100 ->  45°
	0,	// 1101
	0,	// 1110
	0	// 1111
	};

void main(void)
{
	// Oscillator and Clock
	OSCCON = 0x7A;      	// select internal RC 16MHz

	// Fixed Voltage Reference
	FVRCON = 0x80 + (2 << 2) + 3;	// FVR2 = 2V, FVR1 = 4V

	// setup EUSART to send serial data
	SPEN = 1;
	SPBRGL = 25;			// 9600 bit/s @ 16MHz fOSC
	TXEN = 1;

	// setup MSSP to work as SPI master
	SSPADD = 17;			// fOSC / (4 * fSPI), correction for start delay!
	SSPCON1 = 0x2A;         // enable as SPI master
	SMP = 1;                // sample at end
	SSPBUF = 0x55;			// start transfer

	// configure Pins and Ports (A4 and A5 when no external quartz neded)
	//  	 5		   4		  3      	  2			   1		   0
	// A:	SDO		CLKOUT		 nMCLR		  TX		ICSPCLK		ICSPDAT
	// C:	SDI		OPA2IN- 	OPA2OUT		OPA1OUT		OPA1IN-		  AN4
	WPUC = 0b100000;		// C5 with pullup
	nWPUEN = 0;				// allow weak pullup
	ANSC5 = 0;              // digital input
	SSPDATPPS = 0x15;		// SDI from RC5
	RA5PPS = 0x12;          // RA5 from SDO
	RA2PPS = 0x14;          // RA2 from TX
	TRISA = 0b011011;		// RA5 and RA2 are digital outputs

	// setup A/D-Converter
	ADCON0 = (4 << 2) + 1; 	// activate using AN4
	ADCON1 = 0b00100011;	// fOSC / 32, Vref = FVR1, left justified

	// configure Op Amps
	OPA1CON = 0b11000011;	// enable inverting OpAmp
	OPA2CON = 0b11000011;	// enable inverting OpAmp

	// test loop if requested
	while (RC5 == 0) {      // when test mode requested
		while (BF == 0); 	// every 38µs
		spirx = SSPBUF;
		SSPBUF = 0x55;
		level += ADRESH;
		ADGO = 1;      		// next conversion
		if (++loops == 0) {
			TXREG = level >> 8;	// send measured value every 9.7ms
			level = 0;
		}
	}

	// activate comperator1: C1IN2- @ C2 -> C1OUT @ C5
	CM1CON0 = 0b10001110;	// with filter and hysteresis enabled
	CM1CON1 = (6 << 3) + 2;	// C1IN+ = FVR2
	RC5PPS = 0x16;			// RC5 from C1OUT
	TRISC5 = 0;				// RC5 is output now

	// Main Loop
	while (1) {
		while (BF == 0); 	// every 38µs
		spirx = SSPBUF;
		SSPBUF = spitx;
		level += ADRESH;
		ADGO = 1;      		// next conversion
		askval = level >> 4;
		level -= askval;

		// analyze data read from HW blocks
		if (phasetab[spirx & 0x0F] == 0)
		{					// feedback signal incorrect
			if (carrier > 0) carrier--;
			if (carrier < CARR_LOST) {
				if (detectstep != IDLE) TXREG = 0;	// report IDLE state
				detectstep = IDLE;
			}
		}
		else {				// correct feedback
			if (carrier < CARR_STAB) carrier++;
			else if (detectstep == IDLE) {
				TXREG = (askval - 100) | 0x80;		// report ASK value
				detectstep = STAB;
			}
		}
		
		// debugging support
		spitx = (spirx & 0x0F) | 0xC0;
//		testbuf[loops++] = spirx;
	}
}
